#!/usr/local/bin/perl5 -w

# Concept test monitor.  Just display current statistics in the form
# of a bar chart.

# For teacher to monitor the results of a test in progress.
# No authentication here to check teacher.

# FMC
# Ideally, want to be incremental.
#
# Also, want to display original test along with correct answers and
# statistics.

use strict;

# Note Push
#fmc
#use CGI::Push ':standard', ':html3';
use CGI ':standard', ':html3';
use CGI::Carp 'fatalsToBrowser';

use POSIX;
use Fcntl ':flock';
use Storable;
use MLDBM qw(DB_File Storable);
#use DB_File;

# FMC better place?
use lib '.';
use TestBase;

# Parameters
use constant MONITOR => 'monitor';
use constant SAVERESET => 'savereset';
use constant REFRESH => 'refresh';
use constant DUMPTITLE => 'dumptitle';
use constant TESTSTATUS => 'teststatus';
use constant NUMANSWERS => 'numanswers';
use constant NUMANSWERS2 => 'numanswers2';

use constant AUTO_REFRESH_ON => 'Auto Refresh On';
use constant AUTO_REFRESH_OFF => 'Auto Refresh Off';

use constant START_TEST => 'Start Test';
use constant STOP_TEST => 'Stop Test';

use constant COURSE => 'course';
use constant TESTNAME => 'testname';

# Single pixel GIF trick
use constant PIXELGIF => '/icons/1p-blue.gif';


### Useful
my @letter =
  ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
   'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z');

sub answer_prefix {
  my ($num) = @_;

  if ($num == 0) {
    '(None)'
  }
  else {
    $letter[$num-1]
  }
}


### Main

if (param(MONITOR)) {
  my $course = param(COURSE);
  my $test_name = param(TESTNAME);

  # Delay in second between page updates.
  # FMC param?
  my $delay = 3;


  # FMC need factor out with testgen.cgi.
  my $course_dir = course_dir($course);
  if (! -d $course_dir) {
    my $title = 'No such course';
    print
      header,
      start_html($title),
      h1($title),
      "No such course: ",
      em($course),
      ".",
      end_html;
    exit 0;
  }

  my $test_dir = test_dir($course, $test_name);
  my $stored = $test_dir . '/' .
    $test_name . TESTSTORED;
  if (! -s $stored) {
    my $title = 'No such test';
    print
      header,
      start_html($title),
      h1($title),
      "No such test for course ",
      em($course),
      ": ",
      em($test_name),
      ".",
      end_html;
    exit 0;
  }

  my $dbfile = $test_dir . '/' .
    $test_name . TESTDB;

  my %db;

  # Periodically refresh with latest results.
  my $do_current = sub {
#    my ($q, $counter) = @_;
    local(*DB_FH);

    my %result;

    if (param(SAVERESET)) {
      # Don't propagate to self URL.
      Delete(SAVERESET);
      param(REFRESH, AUTO_REFRESH_OFF);

      my $obj = retrieve($stored);

      my $num_answers = param(NUMANSWERS) + 0;
      if ($num_answers < 2) {
        # Avoid dumb cases.
        $num_answers = 2;
      }
      my $num_answers2 = param(NUMANSWERS2) + 0;
      if ( $num_answers2 < 2 ) {
	# be smart
	$num_answers2 = 0;
      }
      # Hack:  set the size of the first question.
      my $item = $obj->{items}->[0];
      $#{$item} = $num_answers;
      foreach my $i (1 .. $num_answers) {
        # Empty answer.
        $item->[$i] = ['', 0];
      }
      
      # Another Hack: set the size of the second question
      # (removing question #2 if num_answers2 < 2)
      while( scalar( @{$obj->{items}} ) > 1 ) {
	pop @{$obj->{items}};
      }

      if ( $num_answers2 != 0 ) {
	my @item;
	$#item = $num_answers2;
	$item[0] = ""; # blank question
	foreach my $i (1 .. $num_answers2) {
	  # empty answer
	  $item[$i] = ['',0];
	}

	push @{$obj->{items}}, \@item;
      }

      # Taint issue.
      store($obj, $stored) or
        die "Could not store test to $stored\n";

      # Each item has a question, then answers, hence the number of
      # answers is the length of the item - 1.
      my @num_answers = map { $#{$_} } @{$obj->{items}};

      # FMC See testsubmit.cgi, testgen.cgi.
      my $db = tie(%db, 'MLDBM',
                   $dbfile,
                   O_RDWR)
        or die "$dbfile: $!";
      my $fd = $db->fd;
      open(DB_FH, "<&=$fd") or die "dup $fd: $!";
      flock(DB_FH, LOCK_SH);

      # Save data.
      my $time = localtime;
      $time =~ s{ }{-}g;
      my $dump_title = param(DUMPTITLE);

      # Protect against some malice.
      $dump_title =~s{/}{__}g;

      my $data_file = $time . '-' . $dump_title . DATASUFFIX;

      # Write into test directory.
      open DATA_FH, "> $test_dir/$data_file" or
        die "$data_file: $!\n";

      foreach my $user (sort keys %{$db{USERS()}}) {
        print DATA_FH $user, "\n";
        foreach my $item (@{$db{USERS()}{$user}}) {
          print DATA_FH "@$item", "\n";
        }
        print DATA_FH "\n";
      }
      close DATA_FH;

      $db{TESTID()} = rand;
      $db{STOPPED()} = 1;
      $db{TESTTYPE()} = TEST_CONCEPT;
      $db{NUM_ANSWERS()} = \@num_answers;
      $db{USERS()} = {};
      %result = %db;

      flock(DB_FH, LOCK_UN);
    }
    elsif (my $test_status = param(TESTSTATUS)) {
      if ($test_status eq START_TEST) {
        param(REFRESH, AUTO_REFRESH_ON);
      }
      else {
        param(REFRESH, AUTO_REFRESH_OFF);
      }

      # Don't propagate to self URL.
      Delete(TESTSTATUS);

      my $db = tie(%db, 'MLDBM',
                   $dbfile,
                   O_RDWR)
        or die "Test score database $dbfile: $!";
      my $fd = $db->fd;
      open(DB_FH, "<&=$fd") or die "dup $fd: $!";
      flock(DB_FH, LOCK_SH);

      $db{STOPPED()} = ($test_status eq STOP_TEST);
      # Slurp into memory and process after unlocking.
      %result = %db;

      flock(DB_FH, LOCK_UN);
    }
    else {
      my $db = tie(%db, 'MLDBM',
                   $dbfile,
                   O_RDONLY)
        or die "Test score database $dbfile: $!";
      my $fd = $db->fd;
      open(DB_FH, "<&=$fd") or die "dup $fd: $!";
      flock(DB_FH, LOCK_SH);

      # Slurp into memory and process after unlocking.
      %result = %db;

      flock(DB_FH, LOCK_UN);
    }
    untie %db;

    my $test_type = $result{TESTTYPE()};
  if ($test_type ne TEST_CONCEPT) {
      my $title = 'Not a concept test';
      print
        header,
        start_html($title),
        h1($title),
        "Test ",
        em($test_name),
        " of course ",
        em($course),
        " is not a concept test.",
        end_html;
      exit 0;
  }

    my $output;


#    if (scalar keys %{$result{USERS()}}) {
      my @rows;
      my $length;

      # Array of distributions for each question.
      # Each distribution is an array of counts for each answer, including
      # 0 (no answer).
      my @data;
      my $count = 0;
      foreach my $n (@{$result{NUM_ANSWERS()}}) {
        # Remember space for 0.
        $#{$data[$count++]} = $n;
      }

      foreach my $user (sort keys %{$result{USERS()}}) {
        my $count = 0;
        foreach my $item (@{$result{USERS()}{$user}}) {
          # Only pay attention to latest answer.
          my $latest = $$item[$#{$item}];
          $data[$count++]->[$latest]++;
        }
      }

      my $stritems = gen_chart(\@data);
      my $i = 0;
      
      my $tblstr = "";
      for( $i = 0; $i < scalar(@{$stritems}); $i+=2 ) {
	my $item1;
	my $item2 = "";
	
	$item1 = td({-width => '50%'},
		    $stritems->[$i]);
	
	if ( $i+1 < scalar(@{$stritems}) ) {
	  $item2 = td({-width => '50%'},
		      $stritems->[$i+1]);
	}
	
	$tblstr .= TR({-valign => 'top'},
		      $item1,$item2);
      }

      $output = table({-width => '100%'},
		      $tblstr);

#    }

    # OK, do it.
    my $title = "Results for Course \"$course\", Test \"$test_name\"";
    my $instructions = $result{STOPPED()} ?
      'The current test is over.' :
        ol(
           li('Go to ' .
              tt("http://quiz.cs.cmu.edu/") .
              ' in the Web browser.'),
           li('Click on "generic concept test".'),
           li('Fill in the test form.'),
           li('Press one of the ' . strong('Submit') . ' buttons.'),
           li('You may resubmit your responses as often as desired.')
          );

    my $time = localtime();

    my $refresh_button = '';
    if (param(REFRESH) eq AUTO_REFRESH_ON) {
      my $s = self_url;
      print header(-refresh => "$delay; URL=$s",
                   -pragma => 'no-cache');
      $refresh_button =
        submit(-name => REFRESH,
               -value => AUTO_REFRESH_OFF)
    }
    else {
      print header(-pragma => 'no-cache');
      if (! $result{STOPPED()}) {
        $refresh_button =
          submit(-name => REFRESH,
                 -value => AUTO_REFRESH_ON)
      }
    }

    my $start_stop_button = $result{STOPPED()} ?
      submit(-name => TESTSTATUS,
             -value => START_TEST) :
      submit(-name => TESTSTATUS,
             -value => STOP_TEST);

    # No header.
#    return
    print
      start_html($title),
      h1($title),
      h2("Instructions for Students"),
      $instructions,
      hr,

      h2('Results'),
      $output,

      start_form,

      $refresh_button,
      $start_stop_button,

      hidden(-name => MONITOR),
      hidden(-name => COURSE),
      hidden(-name => TESTNAME),
      hidden(-name => REFRESH),

      hr,
      submit(-name => SAVERESET,
             -value => 'Save data and reset score database'),
      'Number of answers: ',
      textfield(-name => NUMANSWERS,
                -value => $result{NUM_ANSWERS()}[0],
                -size => 5),
      textfield(-name => NUMANSWERS2,
		-value => $result{NUM_ANSWERS()}[1],
		-size => 5),
      hr,
      'Title (optional) for data: ',
      textfield(-name => DUMPTITLE,
                -size => 20),

      end_form,
      hr,
      $time,
      end_html
    };

#fmc
#  do_push(-delay => $delay,
#          -next_page => $do_current);
    &$do_current();
}
else {
  print_monitor();
}
### End main


sub gen_chart {
  my ($data) = @_;

  my $count = 1;
  my @result = map {
    my $item = $_;
    my $rows = gen_rows($item);

    # Triple
#    my @header = ('Answer', 'Count', 'Bar');
#    unshift(@$rows, td(\@header));

    "Question " . $count++ . ":" .
      table(TR($rows))
  } @$data;

  \@result
}


sub gen_rows {
  my ($item) = @_;
  my $answer_num = 0;

  my $sum = sum($item);

  my @rows = map {
    my $count = $_ + 0;

    # Guess layout!
    my $bar = $count == 0 ?
      '' :
        do {
          my $percentage = $count * 100 / $sum;

          img({ -src => PIXELGIF,
                -alt => "$percentage%",
                -height => 20,
                -width => 3 * $percentage
              });
        };

    # Triple
    my @data = (answer_prefix($answer_num++) . '.', $count, $bar);


    td(\@data)
  } @$item;

  # Put unanswered at end.
  my $none = shift @rows;
  push(@rows, $none);

  \@rows
}


# Return sum of all elements in an array (ref).
sub sum {
  my ($item) = @_;
  my $sum = 0;
  foreach my $i (@$item) {
    $sum += $i
  }
  $sum
}


# Fields:  COURSE, TESTNAME
# MONITOR
sub print_monitor () {
  my $title = 'Monitor Concept Test';

  print
    header,
    start_html($title),
    h1($title),
    start_form,

    'Course code:',
    br,
    textfield(-name => COURSE,
              -size => 20),
    p,

    'Test:',
    br,
    textfield(-name => TESTNAME,
              -size => 20),
    p,

    submit(-name => MONITOR,
           -value => 'Monitor Concept Test'),

    end_form,
    end_html;
}
