#!/usr/local/bin/perl5 -w

# Register an ID for class.

use strict;

use CGI ':standard';
use CGI::Carp 'fatalsToBrowser';
use Time::Local 'timegm';

use POSIX;
use Fcntl ':flock';
use MLDBM qw(DB_File Storable);

# FMC better place?
use lib '.';
use TestBase;

# Parameters
use constant REGISTER => 'register';
use constant USER => 'user';
use constant TESTCONCEPT => 'testconcept';
use constant COURSE => 'course';
use constant TESTNAME => 'testname';
use constant LOGIN => 'login';

# Base name of user database.
use constant USERFILE => 'users.db';

### Main

if (param(USER)) {
  # Check valid user string.
  # FMC should beef up to perform more checks, perhaps password.
  my $user = param(USER);
  if (valid_user($user)) {
    # Save information into a file.
    my $dbfile = COURSEDIR . '/' . USERFILE;
    my %db;
    {
      my $db = tie(%db, 'MLDBM',
                   $dbfile,
                   O_CREAT|O_RDWR)
        or die "User database $dbfile: $!";
      my $fd = $db->fd;
      open(DB_FH, "<&=$fd") or die "dup $fd: $!";
      flock(DB_FH, LOCK_SH);

      $db{$user} = $user;

      flock(DB_FH, LOCK_UN);
    }
    untie %db;

    # compute expiration date
    # Jan <= curdate <= May    expire June 1 0:00
    # Jun <= curdate <= Aug    expire Sep 1 0:00
    # Sep <= curdate <= Dec    expire Jan 1 0:00 next year

    my(@MON)=qw/Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec/;
    my(@WDAY) = qw/Sun Mon Tue Wed Thu Fri Sat/;
    my ($sec,$min,$hour,$mday,$mon,$year,@rest) = gmtime(time);
    my $expire;
    if ($mon <= 4) {
      # Jan-May
      $mon = 5;
    } elsif ($mon <= 7) {
      # Jun-Aug
      $mon = 8;
    } else {
      # Sep-Dec
      $mon = 0;
      ++$year;
    }
    my $wday = (gmtime (timegm (0, 0, 0, 1, $mon, $year))) [6];
    $year += 1900;
    $expire = "$WDAY[$wday], 1-$MON[$mon]-$year 00:00:00 GMT";
    print STDERR $expire, "\n";

    my $cookie = cookie (-name=>'user',
                         -value=>$user,
                         -expires=>$expire);
    
    if (param(COURSE) and param(TESTNAME)) {
      print
        redirect(-uri => 'testconcept.cgi?' .
                 LOGIN . '=' . 'external' . '&' .
                 USER . '=' . CGI->escape($user) . '&' .
		 COURSE . '=' . CGI->escape(param(COURSE)) . '&' .
		 TESTNAME . '=' . CGI->escape(param(TESTNAME)) . '&',
                 -cookie => $cookie);

    } elsif( param(TESTCONCEPT) ) {
      print redirect(-uri => 'testconcept.cgi',
		     -cookie => $cookie);

    } else {
      my $title = 'Registration Accepted';
      print
        header (-cookie=>$cookie),
        start_html($title),
        h1($title),
        "The handheld of user ",
        tt($user),
        " has been registered successfully.",
        hr,
        "Return to ",
        a({ -href => "/"},
          " main page"),
        " or ",
	  a({-href => "/cgi-bin/testconcept.cgi"},
	     "take a concept test"),
	".",
        end_html;
    }
  }
  else {
    my $title = 'Invalid User';
    print
      header,
      start_html($title),
      h1($title),
      "Not a valid user ID: ",
      tt($user),
      ".",
      end_html;
  }
}
else {
  print_register();
}

### End main

# Determine if $user is a valid Andrew ID.
sub valid_user {
  my ($user) = @_;

  # Currently just check lexically OK.
  if ($user =~ /^[A-Za-z0-9_-]+$/) {
    1
  }
  else {
    0
  }
}

sub print_register () {
  my $title = 'Log In';

  print
    header(-pragma => 'no-cache'),
    start_html($title),
    h1($title),
    start_form,
    p,
    'Andrew ID (without ',
    tt('+'),
    ' or ',
    tt('@andrew.cmu.edu'),
    '): ',
    p,
    textfield(-name => USER,
              -size => 10),

    br;

  if (param(COURSE) && param(TESTNAME)) {
    print
      hidden(-name => COURSE,
             -value => param(COURSE)),
      hidden(-name => TESTNAME,
             -value => param(TESTNAME));
  }

  if (param(TESTCONCEPT)) {
    print
      hidden(-name => TESTCONCEPT,
	     -value => param(TESTCONCEPT));
  }

  print
    submit(-name => REGISTER,
           -value => 'Register'),

    end_form,
    end_html;
}
