#!/usr/local/bin/perl5 -wT

# For student to run.
#
# Warning: not secure.  Student could take intermediate HTML and hack
# it and cause bad things to happen upon submission of a form.

use strict;

use CGI ':standard';
use CGI::Carp 'fatalsToBrowser';

use POSIX;
use Fcntl ':flock';
use Storable;
use MLDBM qw(DB_File Storable);
#use DB_File;

# FMC better place?
use lib '.';
use TestBase;

# Parameters
use constant LOGIN => 'login';
use constant USER => 'user';
use constant COURSE => 'course';
use constant TESTNAME => 'testname';
use constant SUBMIT => 'submit';



### Main
autoEscape(0);

if (param(LOGIN)) {
  my $submit = param(SUBMIT);
  my $user = param(USER);
  my $course = param(COURSE);
  my $test_name = param(TESTNAME);

  my $course_dir = course_dir($course);
  if (! -d $course_dir) {
    my $title = 'No such course';
    print
      header,
      start_html($title),
      h1($title),
      "No such course: ",
      em($course),
      ".",
      end_html;
    exit 0;
  }

  my $test_dir = test_dir($course, $test_name);
  my $stored = $test_dir . '/' .
    $test_name . TESTSTORED;
  if ((! -d $test_dir) || (! -s $stored)) {
    my $title = 'No such test';
    print
      header,
      start_html($title),
      h1($title),
      "No such test for course ",
      em($course),
      ": ",
      em($test_name),
      ".",
      end_html;
    exit 0;
  }

  # Check type.
  my $test_type;
  my $already;
  my $dbfile = test_dir($course, $test_name) . '/' .
    $test_name . TESTDB;
  my %db;
  {
    my $db = tie(%db, 'MLDBM',
                 $dbfile,
                 O_RDONLY)
      or die "Test score database $dbfile: $!";
    my $fd = $db->fd;
    open(DB_FH, "<&=$fd") or die "dup $fd: $!";
    flock(DB_FH, LOCK_SH);
    $test_type = $db{TESTTYPE()};
    $already = exists $db{USERS()}{$user};
    flock(DB_FH, LOCK_UN);
    close(DB_FH);
  }
  untie %db;

  if ($test_type ne TEST_ORDINARY) {
      my $title = 'Not an ordinary test';
      print
        header,
        start_html($title),
        h1($title),
        "Test ",
        em($test_name),
        " of course ",
        em($course),
        " is not an ordinary test.",
        end_html;
      exit 0;
  }

  # Check if the student is allowed to take the test.
  if (! $submit) {
    if ($already) {
      my $title = 'Test already taken';
      print
        header,
        start_html($title),
        h1($title),
        em($user),
        " has already taken test ",
        em($test_name),
        " of course ",
        em($course),
        ".",
        end_html;
      exit 0;
    }
  }

  my $obj = retrieve($stored);

  if ($submit) {
    print_results($obj,
                  $user,
                  $course,
                  $test_name);
  }
  else {
    # Globals:  LOGIN, USER, COURSE, TESTNAME
    print_all($obj);
  }
}
else {
  print_login();
}

### End main



# print_login()
#
# Fields:  USER, COURSE, TESTNAME, LOGIN
sub print_login () {
  my $title = 'Test Login';

  print
    header,
    start_html($title),
    h1($title),
    start_form,

    'User:',
    br,
    textfield(-name => USER,
              -size => 10,
              -value => cookie(-name=>'user')),
    p,
    br,

    'Course code:',
    br,
    textfield(-name => COURSE,
              -size => 10),
    p,
    br,

    'Test:',
    br,
    textfield(-name => TESTNAME,
              -size => 10),
    p,
    br,

    submit(-name => LOGIN,
           -value => 'Log In'),

    end_form,
    end_html;
}

sub answer_text { my $arg = shift;
  my ($answer, $correct) = @$arg;
  $answer
}

sub answer_correct { my $arg = shift;
  my ($answer, $correct) = @$arg;
  $correct
}

# Return the index of the correct answer for an item.
# FMC Assumes that there is exactly one correct answer.
sub item_correct { my ($question, @answers) = @$_;
  my $count = 1;
  foreach my $answer (@answers) {
    if (answer_correct($answer)) {
      return $count;
    }
    $count++;
  }

  die "Question \"$question\" is invalid and has no correct answer.\n";
}

BEGIN {
  my $base = 'item';

  # Globals: $base
  sub item_name {
    my ($count) = @_;
    $base . $count
  }
}




# Globals:  LOGIN, USER, COURSE, TESTNAME
sub print_all {
  my ($obj) = @_;
  print
    header,
#    start_html($obj->{title}),
#    h1($obj->{title}),
    $obj->{prologue};

  print_form($obj->{items});

  print
    $obj->{epilogue}
#   end_html;
}



sub print_results {
  my ($obj, $user, $course, $test_name) = @_;
  print
    header,
    start_html($obj->{title}),
    h1($obj->{title});

  print_score($obj->{items}, $user, $course, $test_name);

  print
    end_html;
}


# Depends on item name encoding of answers chosen.
sub print_score {
  my ($items, $user, $course, $test_name) = @_;

  my @iota = 1 .. scalar @$items;

  # The correct answers.
  my @correct = map item_correct, @$items;

  # Get all the user responses.
  # Note strings are coerced to integers here.
  # Also, if an item was not answered, the result will be 0.
  my @response = map { param(item_name($_)) + 0} @iota;

  # Boolean list of matches.
  my @score = map { ($response[$_-1] == $correct[$_-1]) + 0 } @iota;

  my $dbfile = test_dir($course, $test_name) . '/' .
    $test_name . TESTDB;
  my %db;
  my $already;
  {
    my $db = tie(%db, 'MLDBM',
                 $dbfile,
                 O_RDWR)
      or die "Test score database $dbfile: $!";
    my $fd = $db->fd;

    open(DB_FH, "<&=$fd") or die "dup $fd: $!";
    flock(DB_FH, LOCK_EX);
    my $u = $db{USERS()};
    $already = exists $u->{$user};
    if (! $already) {
      # FMC do more.
      # Save score for each user.
      $u->{$user} = \@score;
      $db{USERS()} = $u;
    }
    flock(DB_FH, LOCK_UN);
  }
  untie %db;

  if ($already) {
    # Caught multiple attempts late.
    print
      "User ",
      em($user),
      " has already taken test ",
      em($test_name),
      " of course ",
      em($course),
      ".";
    return undef;
  }

  # FMC sample action.
  my @explanation = map {
    $score[$_-1] ? "OK" :
      em("Correct answer: " . $correct[$_-1])
  } @iota;

  print
    p("User: " . em($user)), br,
    p("Course: " . em($course)), br,
    p("Test: " . em($test_name)), br;

  print_marked_items($items, \@response);

#    ol(
#       li(\@explanation)
#      );

}


# Print entire form.
#
# Want login data to persist.
# Globals:  LOGIN, USER, COURSE, TESTNAME
# SUBMIT
sub print_form { my ($items) = @_;
  print
    start_form,

    # Propagate all state.
    # FMC should make automatic
    hidden(-name => LOGIN),
    hidden(-name => USER),
    hidden(-name => COURSE),
    hidden(-name => TESTNAME),

    ol(
       li(
          gen_items($items)
         )
      ),
    submit(-name => SUBMIT,
           -value => 'Score Test'),
    end_form;
}

#fmc
# Print the questions along with marking of correct and incorrect answers.
sub print_marked_items {
  my ($items, $response) = @_;

  my @item_output;

  my $i = 0;
  foreach my $item (@$items) {
    my ($question, @answers) = @$item;

    # 1-based index.
    my $user_answer = $$response[$i];

    my @answer_output;
    my $j = 1;
    foreach my $answer (@answers) {
      if (answer_correct($answer)) {
        push(@answer_output, strong("(Correct) " . answer_text($answer)));
      }
      elsif ($j == $user_answer) {
        push(@answer_output, em("(You chose incorrectly) ") . answer_text($answer));
      }
      else {
        push(@answer_output, answer_text($answer));
      }
      $j++;
    }

    push(@item_output, $question . li(\@answer_output));
    $i++;
  }

  print
    ol(
       \@item_output
      )
}


# Each item is named by 1-based index into the list of items.
# Each answer has as value the 1-based index into the list of answers.
sub gen_items {
  my ($items) = @_;

  my $count = 1;
  my @result = map {
    my ($question, @answers) = @$_;

    my $name = item_name($count++);
    my @iota = 1 .. scalar @answers;

    $question .
      radio_group(
                  -name => $name,
                  -values => \@iota,
                  -labels => gen_labels(\@answers),
                  -default => '-',
                  -cols => 1
                 )
    } @$items;

  \@result
}


BEGIN {
  my @letter =
    ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
     'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z');

  sub answer_prefix {
    my ($num) = @_;

    if ($num == 0) {
      '(None)'
    }
    else {
      $letter[$num-1]
    }
  }
}

# Generate the labels mapping for the answers of a given item.
sub gen_labels { my ($answers) = @_;
  my $count = 1;
  my %result = map {
    my $c = $count++;
    ($c => answer_prefix($c) . '. ' . answer_text $_)
  } @$answers;

  \%result
}

__END__
